//
//  P4Map.h
//  P4Menu
//
//  Created by Michael Bishop on 12/13/11.
//  Copyright (c) 2011 Numerical Garden LLC. All rights reserved.
//

#import <Foundation/Foundation.h>

/*----------------------------*
 |                            |
 |          P4Mapper          |
 |                            |
 *----------------------------*/

typedef enum P4MappingDirection {
    P4MappingDirectionLeftToRight = 0,
    P4MappingDirectionRightToLeft
} P4MappingDirection;


@interface P4Mapper : NSObject {
    void * _api;
    NSArray * _entries;
}

@property (nonatomic,readwrite,copy) NSArray * entries;

-(NSString*)pathByMappingPath:(NSString*)path direction:(P4MappingDirection)direction;

-(P4Mapper*)mapperByJoiningWithMapper:(P4Mapper*)mapper;
-(P4Mapper*)mapperByJoiningInDirection:(P4MappingDirection)direction
                            withMapper:(P4Mapper*)rightMapper
                           inDirection:(P4MappingDirection)rightDirection;

@end

/*----------------------------*
 |                            |
 |         P4MapEntry         |
 |                            |
 *----------------------------*/


typedef enum P4MapEntryType {
    P4MapEntryTypeInclude = 0,
    P4MapEntryTypeExclude,
    P4MapEntryTypeOverlay
} P4MapEntryType;

@interface P4MapEntry : NSObject {
    P4MapEntryType _type;
    NSString * _leftPath;
    NSString * _rightPath;
}

+(P4MapEntry*)entryWithEntryString:(NSString*)entryString;

+(P4MapEntry*)entryWithLeftPath:(NSString*)leftPath;
+(P4MapEntry*)entryWithLeftPath:(NSString*)leftPath type:(P4MapEntryType)type;

+(P4MapEntry*)entryWithLeftPath:(NSString*)leftPath rightPath:(NSString*)rightPath;
+(P4MapEntry*)entryWithLeftPath:(NSString*)leftPath rightPath:(NSString*)rightPath type:(P4MapEntryType)type;

@property (nonatomic, readwrite, assign) P4MapEntryType type;
@property (nonatomic, readwrite, copy) NSString * leftPath;
@property (nonatomic, readwrite, copy) NSString * rightPath;

@end
