//
//  P4Port.h
//  P4ObjectLayer
//
//  Created by Michael Bishop on 9/1/10.
//  Copyright 2010 Numerical Garden, LLC. All rights reserved.
//

#import <Cocoa/Cocoa.h>
#import "P4RawConnection.h"

typedef void (^PortRefreshCompletionBlock)(NSError*);

@class NGAReachability;
@class P4Response;

@interface P4Port : NSObject {

@private
    NSMutableArray *    _completionBlocksForRefresh;
    BOOL                _isRefreshing;

    NSDictionary *      _info;
    NSString *          _portString;

    NGAReachability *   _reachability;
    BOOL                _reachable;

    NSString *          _GMTOffsetString;
    NSError *           _lastError;
}

+(P4Port*)portWithPortString:(NSString*)p4port;

-(BOOL)runArguments:(NSArray*)arguments
        withContext:(NSDictionary*)context
        updateBlock:(UpdateBlock)update
    completionBlock:(void(^)(P4Response*))completion;

-(BOOL)runArguments:(NSArray*)arguments
        withContext:(NSDictionary*)context
            content:(NSString*)content
        updateBlock:(UpdateBlock)update
    completionBlock:(void(^)(P4Response*))completion;

-(P4Response*)runArguments:(NSArray*)arguments
               withContext:(NSDictionary*)context
                   content:(NSString*)content;

-(IBAction)refresh:(id)sender;

-(void)refreshWithCompletion:(PortRefreshCompletionBlock)completion;

// example: "-0400" in "2010/09/01 13:37:19 -0400 EDT"
// You can append this to server dates to put them in the right time zone
@property (nonatomic,readonly,copy)   NSString *  GMTOffsetString;

@property (nonatomic,readonly,copy)   NSDictionary *  info;
@property (nonatomic,readonly,copy)   NSString     *  portString;
@property (nonatomic,readonly,assign) BOOL            reachable;
@property (readonly,assign) BOOL            refreshing;
@property (nonatomic,readonly,retain) NSError *       lastError;
@end
